/*
#
# Main interface tool for GOPchop.  Implements multiple utility functions.
#
# $Id: Main.cpp,v 1.49 2004/02/10 06:28:29 nemies Exp $
#
# Copyright (C) 2001-2003 Kees Cook
# kees@outflux.net, http://outflux.net/
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# http://www.gnu.org/copyleft/gpl.html
#
*/
/*
 * Initial main.c file generated by Glade. Edit as required.
 * Glade will not overwrite this file.
 */

#include "GOPchop.h"

#include <stdio.h>
#include <inttypes.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/uio.h>
#include <string.h>
#include <errno.h>
#include <time.h>

#include "Vector.h"
#include "List.h"
#include "ElementStream.h"
#include "Pack.h"
#include "GroupOfPictures.h"
#include "Picture.h"
#include "MPEG2Parser.h"
#include "mpeg2consts.h"
#include "widgets.h"

#include "GOPutils.h"

#ifdef __cplusplus
extern "C" {
#endif

/* libmpeg2 */
#include "video_out.h"
#include <mpeg2dec/mpeg2.h>

#ifdef __cplusplus
}
#endif

#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <X11/Xlib.h>
#include <math.h>

#include "main.h"

#include <getopt.h>

/* */
/* globals for everyone to beat on */
/* */
// strings
char *GOPchop_cvsid = "$Id: Main.cpp,v 1.49 2004/02/10 06:28:29 nemies Exp $";
char *GOPchop_version = VERSION;

// objects
MPEG2Parser *mpeg2parser;

// command line options
char * opt_videoout    = NULL; // default to the first libvo driver
char * opt_pipe        = NULL; // default to no command
int    opt_show_states = 0;    // default to not reporting libmpeg2 states

#define GOPCHOP_OUTPUT_LIBVO 0
#define GOPCHOP_OUTPUT_PIPE  1

// decode variables
int                   desired_output = GOPCHOP_OUTPUT_LIBVO;
static mpeg2dec_t    *mpeg2dec = NULL;
static vo_open_t     *output_open = NULL;
static vo_instance_t *output = NULL;

// mark settings
gint main_clist_count = 0;
uint32_t mark_start = 0;
uint32_t mark_end = 0;

// pipe variables
int pipes[2];
int *client_pipe = NULL;

/* global preferences */
/*
 * add new options below.
 * If you want it saved to the rc file, add it to the "parsable_items" list.
 * Update the "handle_rc" function below for rc options.
 * Update the menu activation handler to auto-save the rc file.
 */
global_options options={
    run_loop:             0,
    run_speed:            1,
    auto_refresh:         1,
    ignore_errors:        0,
    drop_orphaned_frames: 1,
};
rc_parse_item parsable_items[] = {
    { "run_loop",             &options.run_loop,             RC_TYPE_BOOLEAN },
    { "run_speed",            &options.run_speed,            RC_TYPE_INTEGER },
    { "auto_refresh",         &options.auto_refresh,         RC_TYPE_BOOLEAN },
    { "ignore_errors",        &options.ignore_errors,        RC_TYPE_BOOLEAN },
    { "drop_orphaned_frames", &options.drop_orphaned_frames, RC_TYPE_BOOLEAN },
    { NULL, NULL, 0 }
};

///// Callbacks that glade doesn't know about
void
on_GOP_selector_value_changed(GtkAdjustment * adjustment, gpointer user_data)
{
    uint32_t GOP;

    GOP = (uint32_t) GTK_ADJUSTMENT(adjustment)->value;

    // only do something if we have output, and auto refreshing
    if (options.auto_refresh)
        show_GOP(GOP);

    // make sure the parser is loaded
    if (mpeg2parser->getParsingState()==PARSER_STATE_FINISHED)
        update_GOP_info(GOP);

    // handle start/end button availability
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button_start_mark))) {
        gtk_widget_set_sensitive(button_end_mark, (GOP >= mark_start));
    }
}

void
on_slider_run_speed_value_changed(GtkAdjustment * adjustment, gpointer user_data)
{

    options.run_speed = (int) GTK_ADJUSTMENT(adjustment)->value;
    rc_save(PACKAGE,parsable_items);
}

///// My various supporting functions
void flush()
{
    while (g_main_iteration(FALSE));
    /*
       while(gtk_events_pending())
       gtk_main_iteration();
     */
}

void status_on(const gchar * owner, char *text)
{
    guint context;

    context = gtk_statusbar_get_context_id(GTK_STATUSBAR(main_statusbar),
                                           owner);
    gtk_statusbar_push(GTK_STATUSBAR(main_statusbar), context, text);
    flush();
}

void status_off(const gchar * owner)
{
    guint context;

    context = gtk_statusbar_get_context_id(GTK_STATUSBAR(main_statusbar),
                                           owner);
    gtk_statusbar_pop(GTK_STATUSBAR(main_statusbar), context);
    flush();
    //gtk_main_iteration_do(0);

}

const gchar *fileops = "fileops";

void progress(char *task, float done)
{
    static gchar *lasttask = NULL;
    static float lastdone = 0.0;
    float diff;

    int tasksdiffer;
    gint context;

    context = gtk_statusbar_get_context_id(GTK_STATUSBAR(main_statusbar),
                                           fileops);

    if (lasttask)
    {
        tasksdiffer = strcmp(lasttask, task);
    }
    else
    {
        tasksdiffer = 1;
    }

    if (tasksdiffer)
    {
        if (lasttask)
            gtk_statusbar_pop(GTK_STATUSBAR(main_statusbar), context);

        lasttask = (char *) g_realloc(lasttask, strlen(task) + 1);
        strcpy(lasttask, task);

        gtk_statusbar_push(GTK_STATUSBAR(main_statusbar), context, lasttask);
        flush();
    }

    diff = fabs(done - lastdone);

    if (!tasksdiffer && diff < 0.01)
        return;

    lastdone = done;

    gtk_progress_set_percentage(GTK_PROGRESS(main_progressbar), lastdone);

    /*   sending to stderr instead...
       if (mpeg2parser->getError())
       show_error(mpeg2parser->getError());
     */

    flush();
}

void show_error(char *text)
{
    GtkTextIter iter;
    GtkTextBuffer *buffer;
    static gchar *str = NULL;

    // text needs a static area, I think... ?
    (void *) str = g_realloc(str, strlen(text) + 1);
    strcpy(str, text);

    /*
    gtk_text_freeze(GTK_TEXT(error_text_why));
    gtk_text_set_point(GTK_TEXT(error_text_why), 0);
    gtk_text_forward_delete(GTK_TEXT(error_text_why),
                            gtk_text_get_length(GTK_TEXT(error_text_why)));
    gtk_text_insert(GTK_TEXT(error_text_why), NULL, NULL, NULL,
                    (str), strlen(str));
    gtk_text_set_point(GTK_TEXT(error_text_why),
                       gtk_text_get_length(GTK_TEXT(error_text_why)));
    gtk_text_thaw(GTK_TEXT(error_text_why));

    gtk_widget_show(GTK_WIDGET(error_dialog));
    */


    buffer=gtk_text_view_get_buffer(GTK_TEXT_VIEW(error_text_why));
    gtk_text_buffer_get_end_iter(GTK_TEXT_BUFFER(buffer),&iter);
    gtk_text_iter_forward_to_end(&iter);
    gtk_text_buffer_insert(buffer,&iter,text,0);

}


void get_clips_from_list(GtkTreeModel *model,
                         GtkTreeIter  *iter,
                         struct clip_list_t * clips)
{
    GValue list_GOP_first;
    GValue list_GOP_last;

    if (!model || !iter || !clips) return;

    memset(&list_GOP_first,0,sizeof(list_GOP_first));
    memset(&list_GOP_last, 0,sizeof(list_GOP_last));

    gtk_tree_model_get_value(GTK_TREE_MODEL(model), iter,
                             ITEM_GOP_FIRST, &list_GOP_first);
    gtk_tree_model_get_value(GTK_TREE_MODEL(model), iter,
                             ITEM_GOP_LAST,  &list_GOP_last);

    clips->GOP_first = g_value_get_uint(&list_GOP_first);
    clips->GOP_last =  g_value_get_uint(&list_GOP_last);

    g_value_unset(&list_GOP_first);
    g_value_unset(&list_GOP_last);
}


/* "foreach" functions used to walk the main_clist */
gboolean foreach_count_GOPs(GtkTreeModel *model,
                            GtkTreePath *path,
                            GtkTreeIter *iter,
                            gpointer data)
{
    uint32_t * count = (uint32_t *)data;
    struct clip_list_t clips;

    if (!count) return TRUE;

    get_clips_from_list(model,iter,&clips);

    *count += clips.GOP_last - clips.GOP_first + 1;

    return FALSE; /* keep traversing */
}

struct save_clip_info_t {
    FILE       *file;
    uint32_t    GOP_previous;
    uint32_t    count;
    uint32_t    total;
    off_t       written_bytes;
    char       *progress_task;
};

gboolean foreach_save_clips_to_file(GtkTreeModel *model,
                                    GtkTreePath *path,
                                    GtkTreeIter *iter,
                                    gpointer data)
{
    struct clip_list_t clips;
    uint32_t i;
    struct save_clip_info_t * clip_info = (struct save_clip_info_t*)data;

    if (!clip_info) return TRUE; /* abort */

    get_clips_from_list(model,iter,&clips);


    for (i = clips.GOP_first; i <= clips.GOP_last; i++) {
        progress(clip_info->progress_task,
                 (float)( (float) clip_info->count++ /
                          (float) clip_info->total     ) );

        clip_info->written_bytes +=
            write_GOP(clip_info->file, mpeg2parser, i,
                      ((clip_info->GOP_previous + 1) == i),
                      options.drop_orphaned_frames);
        clip_info->GOP_previous = i;
    }

    return FALSE; /* keep on steppin */
}

void save_file(char *filename)
{
    char buf[1024];
    FILE *mpeg2;
    struct save_clip_info_t clip_info;

    clip_info.progress_task = _("Saving MPEG2");

    if (!(mpeg2 = fopen(filename, "w")))
    {
        snprintf(buf, 1024, "%s: %s\n", filename, strerror(errno));
        show_error(buf);
    }
    else
    {
        time_t begin, finish;
        char report[REPORT_LENGTH];

        time(&begin);
        clip_info.total=0;

        /* count up how many GOPs there are to write */
        gtk_tree_model_foreach(GTK_TREE_MODEL(main_list_store),
                               foreach_count_GOPs,
                               &clip_info.total);

        clip_info.file=mpeg2;
        clip_info.count=0;
        clip_info.written_bytes=0;
        // make sure the first GOP is closed
        clip_info.GOP_previous = (uint32_t) -2;
        /* save each clip in the list */
        gtk_tree_model_foreach(GTK_TREE_MODEL(main_list_store),
                               foreach_save_clips_to_file,
                               &clip_info);

        progress(clip_info.progress_task, 1.0);
        status_off(fileops);

        // write an end-of-PS marker
        fwrite(eos_buf, sizeof(eos_buf), 1, mpeg2);
        fclose(mpeg2);

        time(&finish);
        finish -= begin;

        clip_info.written_bytes /= 1048576;
        if (finish > 0)
            clip_info.written_bytes /= finish;
        if (clip_info.written_bytes == 0)
            clip_info.written_bytes = 1;

        // FIXME: bad pluralization
        snprintf(report,REPORT_LENGTH,finish == 1
                ? _("Wrote %s (GOPs: %u) in %d second (%%%sMB/s)\n")
                : _("Wrote %s (GOPs: %u) in %d seconds (%%%sMB/s)\n"),
                filename, clip_info.total, finish, OFF_T_FORMAT);

        fprintf(stderr, report, clip_info.written_bytes);
    }
}

void open_file(char *filename)
{
    time_t begin, finish;
    char buf[128];
    off_t speed;
    char report[REPORT_LENGTH];
    char * gop_cache_filename;

    time(&begin);
    if (!mpeg2parser->init(filename, progress))
    {
        show_error(mpeg2parser->getError());
    }
    else
    {
        mpeg2parser->parse(options.ignore_errors);
        if (mpeg2parser->getError())
            show_error(mpeg2parser->getError());

        // make sure we actually HAVE something...
        List *GOPs = mpeg2parser->getGOPs();
        if (GOPs && GOPs->getNum())
            file_is_loaded();
        else
            mpeg2parser->reset();
    }

    // load-time calculation
    time(&finish);
    finish -= begin;
    speed = mpeg2parser->getSize() / 1048576;   // max speed is max size
    if (finish > 0)
        speed /= finish;
    if (speed == 0)
        speed = 1;              // claim at least 1MB/s

    snprintf(report,REPORT_LENGTH,finish == 1
            ? _("Loaded %s in %d second (%%%sMB/s)\n")
            : _("Loaded %s in %d seconds (%%%sMB/s)\n"),
            filename, finish, OFF_T_FORMAT);

    fprintf(stderr, report, speed);

    progress(_("Caching"), 1.0);
    /* write out the offset map */
    if (!(gop_cache_filename=(char*)malloc(strlen(filename)+5))) {
        perror("malloc");
    }
    else {
        sprintf(gop_cache_filename,"%s.gop",filename);
        if (write_gop_cache(gop_cache_filename,mpeg2parser)) {
            printf("%s",_("Skipping offset cache writing.\n"));
        }
    }

    progress(_("Done"), 1.0);
    status_off(fileops);
}

/* gtk stupid: no way to get row count from a clist */
void update_GOP_slice_count(int count) {
    gchar buf[128];

    /* update the count of clips */
    sprintf(buf, _("Clips: %d"), count);
    gtk_label_set_text(GTK_LABEL(main_label_clips), buf);
}

void add_GOP_slice(uint32_t start, uint32_t end)
{
    GtkTreeIter iter;
    gchar *pathname;

    pathname = mpeg2parser->getFilename();
    if (!pathname)
        pathname = _("unknown pathname");

    gtk_list_store_append (main_list_store, &iter);
    gtk_list_store_set (main_list_store, &iter,
                        ITEM_FILENAME,  pathname,
                        ITEM_GOP_FIRST, start,
                        ITEM_GOP_LAST,  end,
                        -1);

    main_clist_count++;
    update_GOP_slice_count(main_clist_count);
}

void clear_GOP_slices()
{
    gchar buf[128];

    gtk_list_store_clear(main_list_store);

    main_clist_count=0;
    update_GOP_slice_count(main_clist_count);
}

void remove_GOP_slice(GtkTreeIter *iter)
{
    if (!iter) return;

    gtk_list_store_remove(main_list_store,iter);

    main_clist_count--;
    update_GOP_slice_count(main_clist_count);
}

GdkFilterReturn filter_event(GdkXEvent *xevent,
                             GdkEvent  *event,
                             gpointer   data)
{
    //report_event(event);
    printf("got event\n");
    return GDK_FILTER_CONTINUE;
}

GdkWindow          *frame_window = NULL;
const mpeg2_info_t *parser_info = NULL;

void frame_draw()
{
    if (output && output->draw &&
        parser_info && parser_info->display_fbuf) {

        output->draw(output,
                     parser_info->display_fbuf->buf,
                     parser_info->display_fbuf->id);
    }
}

// decode functions
void decode(uint8_t * buffer, size_t len)
{
    Window xwindow;
    vo_setup_result_t setup_result;
    int state;
#ifdef DEBUG_VES
    FILE *ves_debug;
#endif

    //printf("decoding %ld bytes of VES data\n",len);

    /* make sure output and decoder are available */
    if (!output)
    {
        fprintf(stderr, "%s", _("Output window missing!\n"));
        return;
    }
    if (!mpeg2dec)
    {
        fprintf(stderr, "%s", _("MPEG2 decoder missing!\n"));
        return;
    }

    if (!buffer) {
        fprintf(stderr,"%s",_("decode called with a missing buffer!?\n"));
        return;
    }

    //if (len) {
        mpeg2_buffer(mpeg2dec, buffer, buffer + len);
        if (!(parser_info = mpeg2_info(mpeg2dec)))
        {
            fprintf(stderr, "%s", _("MPEG2 decoder info not available!?\n"));
            exit(1);
        }
    //}

#ifdef DEBUG_VES
    /* dump the VES packets to a file */
    ves_debug = fopen("video.es", "a");
    fwrite(buffer, len, 1, ves_debug);
    fclose(ves_debug);
#endif

    while ((state = mpeg2_parse(mpeg2dec)) != STATE_BUFFER)
    {
        switch (state)
        {
            case STATE_SEQUENCE:
                if (opt_show_states) printf("STATE_SEQUENCE\n");
                if (!parser_info->sequence) {
                    fprintf(stderr,"%s",_("Got STATE_SEQUENCE without a sequence!?\n"));
                    break;
                }
                /*
                printf("saw sequence (%dx%d chroma: %dx%d)\n",
                                  parser_info->sequence->width,
                                  parser_info->sequence->height,
                                  parser_info->sequence->chroma_width,
                                  parser_info->sequence->chroma_height
                        );
                */
                /* sequence started, so initialize output window */
                if (output->setup(output,
                                  parser_info->sequence->width,
                                  parser_info->sequence->height,
                                  parser_info->sequence->chroma_width,
                                  parser_info->sequence->chroma_height,
                                  &setup_result))
                {
                    fprintf(stderr, "%s", _("Failed to setup display\n"));
                    return;
                }

                if (output->window_handle) {
                    xwindow = (Window)output->window_handle(output);
                    frame_window = gdk_window_foreign_new((int)xwindow);
                    if (frame_window) {
                        gdk_window_set_events(frame_window,GDK_ALL_EVENTS_MASK);
                        XSelectInput(GDK_DISPLAY(), xwindow,
                                     VisibilityChangeMask
                                     // | PointerMotionMask
                                     );
                    }
                }

                /* does the output need to be converted? */
                if (setup_result.convert)
                    mpeg2_convert(mpeg2dec, setup_result.convert, NULL);
                /* can the output provide an fbuf area? */
                if (output->set_fbuf)
                {
                    uint8_t *buf[3];
                    void *id;

                    mpeg2_custom_fbuf(mpeg2dec, 1);

                    output->set_fbuf(output, buf, &id);
                    mpeg2_set_buf(mpeg2dec, buf, id);

                    output->set_fbuf(output, buf, &id);
                    mpeg2_set_buf(mpeg2dec, buf, id);
                }
                else if (output->setup_fbuf)
                {
                    uint8_t *buf[3];
                    void *id;

                    output->setup_fbuf(output, buf, &id);
                    mpeg2_set_buf(mpeg2dec, buf, id);
                    output->setup_fbuf(output, buf, &id);
                    mpeg2_set_buf(mpeg2dec, buf, id);
                    output->setup_fbuf(output, buf, &id);
                    mpeg2_set_buf(mpeg2dec, buf, id);
                }
                break;
            case STATE_PICTURE:
                if (opt_show_states) {
                    printf("STATE_PICTURE:   ");
                    if (parser_info &&
                        parser_info->current_picture) {
                        switch (parser_info->current_picture->flags &
                                PIC_MASK_CODING_TYPE) {
                            case PIC_FLAG_CODING_TYPE_I:
                                printf("I");
                                break;
                            case PIC_FLAG_CODING_TYPE_B:
                                printf("B");
                                break;
                            case PIC_FLAG_CODING_TYPE_P:
                                printf("P");
                                break;
                            case PIC_FLAG_CODING_TYPE_D:
                                printf("D");
                                break;
                            default:
                                printf("unknown");
                                break;
                        }
                        printf("(%02d)",
                               parser_info->current_picture->temporal_reference);
                    }
                    printf("       parsed\n");
                }

                /* picture is starting */
                if (output->set_fbuf)
                {
                    uint8_t *buf[3];
                    void *id;

                    output->set_fbuf(output, buf, &id);
                    mpeg2_set_buf(mpeg2dec, buf, id);
                }
                if (output->start_fbuf) {
                    if (parser_info->current_fbuf) {
                        output->start_fbuf(output,
                                           parser_info->current_fbuf->buf,
                                           parser_info->current_fbuf->id);
                    }
                    else {
                        fprintf(stderr,"%s",_("Got STATE_PICTURE without a current_fbuf!?\n"));
                    }
                }
                break;
            case STATE_SLICE:
            case STATE_FLUSHED:
            case STATE_END:
                if (opt_show_states) {
                    printf("STATE_%s       ",
                            state == STATE_SLICE ?    "SLICE:    "
                              : (state == STATE_END ? "END:      "
                                 :                    "FLUSHED:  "));
                    if (parser_info &&
                        parser_info->display_picture)
                    {
                        switch (parser_info->display_picture->flags & PIC_MASK_CODING_TYPE)
                        {
                            case PIC_FLAG_CODING_TYPE_I:
                                printf("I");
                                break;
                            case PIC_FLAG_CODING_TYPE_P:
                                printf("P");
                                break;
                            case PIC_FLAG_CODING_TYPE_B:
                                printf("B");
                                break;
                            case PIC_FLAG_CODING_TYPE_D:
                                printf("D");
                                break;
                            default:
                                printf("unknown");
                                break;
                        }
                        printf("(%02d)",
                           parser_info->display_picture->temporal_reference);
                    }
                    else
                    {
                        printf("      nothing loaded to be");
                    }
                    printf(" rendered\n");
                }

                /* draw current picture */
                frame_draw();
                if (output->discard && parser_info->discard_fbuf)
                    output->discard(output,
                                    parser_info->discard_fbuf->buf,
                                    parser_info->discard_fbuf->id);
                break;
            case STATE_INVALID:
                if (opt_show_states) printf("STATE_INVALID\n");
                break;
            case STATE_INVALID_END:
                if (opt_show_states) printf("STATE_INVALID_END\n");
                break;
            case STATE_SEQUENCE_REPEATED:
                if (opt_show_states) printf("STATE_SEQUENCE_REPEATED\n");
                break;
            case STATE_GOP:
                if (opt_show_states) printf("STATE_GOP\n");
                break;
            case STATE_SLICE_1ST:
                if (opt_show_states) printf("STATE_SLICE_1ST\n");
                break;
            case STATE_PICTURE_2ND:
                if (opt_show_states) printf("STATE_PICTURE_2ND\n");
                break;
            default:
                if (opt_show_states) printf("STATE: %d\n",state);
                break;
        }
    }
    if (opt_show_states) printf("STATE_BUFFER\n");
}

void decode_stop()
{
    //printf("decode stop\n");

    if (output)
    {
        // invalidate these
        frame_window = NULL;
        parser_info  = NULL;

        mpeg2_close(mpeg2dec);
        mpeg2dec = NULL;
        if (output->close)
            output->close(output);
        output = NULL;
        output_open = NULL;
    }

    if (client_pipe)
    {
        close(*client_pipe);
        client_pipe = NULL;
    }
}

vo_open_t * find_libvo_driver(char * desired_driver)
{
    int i;
    vo_driver_t const *drivers;

    drivers = vo_drivers();
    for (i = 0; drivers[i].name; i++)
        if (!desired_driver || !strcmp(drivers[i].name, desired_driver))
            return drivers[i].open;

    return NULL;
}


void decode_start()
{
    if (desired_output==GOPCHOP_OUTPUT_LIBVO)
    {

        // decoding
        uint32_t accel = 0;

        //printf("decode start\n");
        if (!(output_open=find_libvo_driver(opt_videoout)))
        {
            fprintf(stderr, _("Can't find output driver '%s'\n"), opt_videoout);
            exit(1);
        }
    
        if (!(output = output_open()))
        {
            fprintf(stderr, _("Can't open output driver '%s'\n"), opt_videoout);
        }
        frame_window = NULL; // invalidate this
    
        accel = mpeg2_accel(MPEG2_ACCEL_DETECT);
#ifdef ARCH_X86
        if (accel & MPEG2_ACCEL_X86_MMXEXT)
            printf("%s", _("Using x86 MMXext acceleration\n"));
        else if (accel & MPEG2_ACCEL_X86_3DNOW)
            printf("%s", _("Using x86 3DNow acceleration\n"));
        else if (accel & MPEG2_ACCEL_X86_MMX)
            printf("%s", _("Using x86 MMX acceleration\n"));
#endif
#ifdef ARCH_PPC
        if (accel & MPEG2_ACCEL_PPC_ALTIVEC)
            printf("%s", _("Using PowerPC Altivec acceleration\n"));
#endif
#ifdef ARCH_ALPHA
        if (accel & MPEG2_ACCEL_ALPHA_MVI)
            printf("%s", _("Using Alpha MVI acceleration\n"));
        else if (accel & MPEG2_ACCEL_ALPHA)
            printf("%s", _("Using Alpha acceleration\n"));
#endif
#ifdef ARCH_SPARC
        if (accel & MPEG2_ACCEL_SPARC_VIS)
            printf("%s", _("Using Sparc VIS acceleration\n"));
        else if (accel & MPEG2_ACCEL_SPARC_VIS2)
            printf("%s", _("Using Sparc VIS2 acceleration\n"));
#endif
        else
            printf("%s", _("Using no special acceleration\n"));
    
        if (!(mpeg2dec = mpeg2_init()))
        {
            fprintf(stderr, "%s", _("Cannot initialize mpeg2 decoder!\n"));
            exit(1);
        }
    }
    else // GOPCHOP_OUTPUT_PIPE
    {
        // try a piped command instead

        if (!opt_pipe)
        {
            fprintf(stderr, "%s", _("Pipe command required.  Try --pipe\n"));
            exit(1);
        }

        // fork off a command
        if (pipe(pipes) < 0)
        {
            perror("pipe");
            exit(1);
        }

        pid_t pid = fork();

        // failure
        if (pid < 0)
        {
            perror("fork");
            exit(1);
        }

        // child
        if (pid == 0)
        {
            char * shell;

            close(pipes[1]);
            if (dup2(pipes[0], STDIN_FILENO) < 0)
            {
                perror("dup2");
                exit(1);
            }
    
            // get the shell to use
            if (!(shell=getenv("SHELL")))
            {
                fprintf(stderr,"%s",
                        _("Cannot figure out from environment what shell to exec.\n"));
                exit(1);
            }

            execl(shell,shell,"-c", opt_pipe, NULL);
            perror("execl");
            exit(1);
        }
    
        // parent
        client_pipe = &pipes[1];
        close(pipes[0]);
    }
}

void clear_GOP_info()
{
    gtk_list_store_clear(GOP_list_store);
    gtk_label_set_text(GTK_LABEL(GOP_label_filename), "");
    gtk_label_set_text(GTK_LABEL(GOP_label_sequence_info),
        _("No Sequence Info Found") );
}

void update_GOP_info(uint32_t num)
{
    int i;
    GroupOfPictures *GOP;
    Bounds *picture_bounds;
    Bounds *packet_bounds;
    Picture *picture;
    List *GOPs;
    List *picture_list;
    List *packet_list;
    List *audio_list;
    List *video_list;
    Vector *head_vector;
    uint32_t picture_index, picture_max;
    uint32_t packet_min, packet_index, packet_max;
    uint32_t audio_index, audio_max;
    off_t gop_len, gop_start;
    off_t audio_len;
    //int   aindex, amax;
    uint8_t *head;
    int drop, hour, min, sec, pictures, closed, broken;

    GtkTreeIter iter;
    gchar datatype[128];
    gchar dataoffset[128];
    gchar datasize[128];

    gchar seqbuf[128];
    struct sequence_info info;

    /* wipe out the old list */
    gtk_list_store_clear(GOP_list_store);
    // wipe out label
    gtk_label_set_text(GTK_LABEL(GOP_label_sequence_info),
                       _("No Sequence Info Found"));

    if (!(GOPs = mpeg2parser->getGOPs()))
    {
        printf(_("no GOPs ?!\n"));
        goto failure;
    }
    if (!(packet_list = mpeg2parser->getPackets()))
    {
        printf(_("NULL picture list ?!\n"));
        goto failure;
    }
    if (!(picture_list = mpeg2parser->getPictures()))
    {
        printf(_("NULL picture list ?!\n"));
        goto failure;
    }
    if (!(audio_list = mpeg2parser->getAudio()))
    {
        printf(_("NULL audio list ?!\n"));
        goto failure;
    }
    if (!(video_list = mpeg2parser->getVideo()))
    {
        printf(_("NULL video list ?!\n"));
        goto failure;
    }
    if (!(GOP = (GroupOfPictures *) GOPs->vector(num)))
    {
        printf(_("no such GOP %d ?!\n"), num);
        goto failure;
    }
    if (!(head_vector = GOP->getHeader()))
    {
        printf(_("GOP header missing?!\n"));
        goto failure;
    }
    if (!(head = mpeg2parser->bytesAvail(head_vector->getStart(), 8)))
    {
        printf(_("GOP header not available?!\n"));
        goto failure;
    }

    // calculate GOP information
    /*
       4         5         6        7
       |         |         |        |
       7 65432 107654 3 210765 432107 6 543210
       1 11111 111111 1 111111 111111 1 1
       d hour  min    m sec    pic    c b
       r              a               l roken
       o              r               osed
       p              k
     */
    drop = ((head[4] & 0x80) > 0);
    hour = ((head[4] & 0x7C) >> 2);
    min = ((head[4] & 0x3) << 4) | ((head[5] & 0xF0) >> 4);

    // sec/picture decoding corrected, care of Scott Smith
    sec = ((head[5] & 0x7) << 3) | ((head[6] & 0xE0) >> 5);
    pictures = ((head[6] & 0x1F) << 1) | ((head[7] & 0x80) >> 7);

    closed = ((head[7] & 0x40) > 0);
    broken = ((head[7] & 0x20) > 0);

    if (!(picture_bounds = GOP->getPictureBounds()))
    {
        printf(_("NULL GOP picture bounds?!\n"));
        goto failure;
    }

    // FIXME: technically, this loop does a lot of needless
    //        reassigning of values.  However, this loop doesn't
    //        happen often, and it's the easiest way I could think
    //        of to sort by offset.

    // get picture bounds
    picture_index = picture_bounds->getFirst();
    picture_max = picture_bounds->getMax();

    for (; picture_index < picture_max; picture_index++)
    {
        Bounds *pic_bounds;
        uint32_t ves, ves_min, ves_max;
        Vector *pic_vector;
        off_t pic_start;
        off_t pic_len;

        if (!(picture = (Picture *) picture_list->vector(picture_index)))
        {
            printf(_("NULL picture?!\n"));
            goto failure;
        }

        if (!(pic_bounds = picture->getVideoBounds()))
        {
            printf(_("NULL picture bounds?!\n"));
            goto failure;
        }
        ves_min = pic_bounds->getFirst();
        ves_max = pic_bounds->getMax();
        pic_len = 0;
        for (ves = ves_min; ves < ves_max; ves++)
        {
            if (!(pic_vector = (Vector *) video_list->vector(ves)))
            {
                printf(_("NULL picture VES vector?!\n"));
                goto failure;
            }
            if (ves == ves_min)
                pic_start = pic_vector->getStart();
            pic_len += pic_vector->getLen();
        }

        snprintf(datatype,   128, _("  Picture (%s: %02d)"),
                frame_type_str(picture->getType()), picture->getTime());
        snprintf(dataoffset, 128, "%" OFF_T_FORMAT, pic_start);
        snprintf(datasize,   128, "%" OFF_T_FORMAT, pic_len);
        gtk_list_store_append(GOP_list_store, &iter);
        gtk_list_store_set(GOP_list_store, &iter,
                ITEM_DATATYPE, datatype,
                ITEM_OFFSET,   dataoffset,
                ITEM_SIZE,     datasize,
                -1);
    }

    // process GOP packets
    if (!(packet_bounds = GOP->getPacketBounds()))
    {
        printf(_("NULL GOP packet bounds?!\n"));
        goto failure;
    }
    packet_min = packet_bounds->getFirst();
    packet_max = packet_bounds->getMax();
    gop_len = 0;
    audio_len = 0;
    for (packet_index = packet_min; packet_index < packet_max; packet_index++)
    {
        Pack *packet;

        if (!(packet = (Pack *) packet_list->vector(packet_index)))
        {
            printf(_("NULL GOP packet vector?!\n"));
            goto failure;
        }
        if (packet_index == packet_min)
            gop_start = packet->getStart();
        gop_len += packet->getLen();

        audio_index = packet->getAudioFirst();
        audio_max = packet->getAudioMax();

        for (; audio_index < audio_max; audio_index++)
        {
            Bounds *audio_bounds;
            int aes, aes_min, aes_max;
            Vector *audio_vector;

            if (!(audio_vector = (Vector *) audio_list->vector(audio_index)))
            {
                printf(_("NULL GOP AES vector?!\n"));
                goto failure;
            }

            audio_len += audio_vector->getLen();

        }
    }

    // display GOP
    snprintf(datatype,   128, _("GOP (%02d:%02d:%02d.%02d%s%s%s)"),
            hour, min, sec, pictures,
            drop ? _(" drop") : _(" keep"),
            closed ? _(" closed") : _(" open"),
            broken ? _(" broken") : _(" whole"));
    snprintf(dataoffset, 128, "%" OFF_T_FORMAT, gop_start);
    snprintf(datasize,   128, "%" OFF_T_FORMAT, gop_len);
    gtk_list_store_prepend(GOP_list_store, &iter);
    gtk_list_store_set(GOP_list_store, &iter,
            ITEM_DATATYPE, datatype,
            ITEM_OFFSET,   dataoffset,
            ITEM_SIZE,     datasize,
            -1);

    // display audio info
    snprintf(datatype,   128, "%s", _("  Audio"));
    snprintf(dataoffset, 128, "%" OFF_T_FORMAT, gop_start);
    snprintf(datasize,   128, "%" OFF_T_FORMAT, audio_len);
    gtk_list_store_append(GOP_list_store, &iter);
    gtk_list_store_set(GOP_list_store, &iter,
            ITEM_DATATYPE, datatype,
            ITEM_OFFSET,   dataoffset,
            ITEM_SIZE,     datasize,
            -1);

    // update labels with info
    GOP->get_sequence_info(&info);    
    snprintf(seqbuf,128,_("%dx%d: %s aspect, %s fps, %s%s"),
        info.horizontal_size, info.vertical_size,
        aspect_str[info.aspect_ratio],
        frame_str[info.frame_rate],
        speed_str(info.bit_rate * 400),
        info.constrained ? _(", constrained") : "");
    gtk_label_set_text(GTK_LABEL(GOP_label_sequence_info),seqbuf);

  failure:
    /* perform any cleanups */
    while (0) {}; // keep compiler quiet
}

#define MAX_FRAME_TYPES 4
const char *frame_type_map[MAX_FRAME_TYPES] = {
    N_("Bad Frame"),
    N_("I-Frame"),
    N_("P-Frame"),
    N_("B-Frame")
};
const char *frame_type_str(int type)
{
    if (type >= MAX_FRAME_TYPES || type < 0)
        type = 0;
    return _(frame_type_map[type]);
}


void show_GOP(uint32_t wantedGOP)
{
    static int initdone = 0;
    List *packets;
    Pack *packet;
    List *GOPs;
    GroupOfPictures *GOP;
    Bounds *bounds;
    ElementStream *ves;
    int start, stop;

    uint8_t *loc;
    size_t len;

    // get our GOP list
    if (!(GOPs = mpeg2parser->getGOPs()))
    {
        printf("%s", _("\tNo GOPs\n"));
        return;
    }

    // get our packet list
    if (!(packets = mpeg2parser->getPackets()))
    {
        printf("%s", _("\tNo packets\n"));
        return;
    }

    // get GOP we're after
    if (!(GOP = (GroupOfPictures *) GOPs->vector(wantedGOP)))
    {
        printf(_("\tGOP %d missing\n"), wantedGOP);
        return;
    }

    // get our GOP's range of packets
    bounds = GOP->getPacketBounds();
    // remember our start and finish points
    start = bounds->getFirst();
    stop = bounds->getMax();

    for (uint32_t j = start; j < stop; j++)
    {
        if (!(packet = (Pack *) packets->vector(j)))
        {
            printf("%s", _("NULL packet?!\n"));
            exit(1);
        }

        if (desired_output==GOPCHOP_OUTPUT_LIBVO && output)
        {
            uint32_t ves_start, ves_stop;

            ves_start = packet->getVideoFirst();
            ves_stop = packet->getVideoMax();

            for (uint32_t k = ves_start; k < ves_stop; k++)
            {
                if (!(ves =
                      (ElementStream *)mpeg2parser->getVideo()->vector(k)))
                {
                    printf("%s", _("NULL VES?!\n"));
                    exit(1);
                }
    
                len = ves->getLen();
                (void *)loc = mpeg2parser->bytesAvail(ves->getStart(), len);
    
                /*
                fprintf(stderr,_("\t\tVES: %d @ %llu (%d): 0x%08x\n"),
                        k,ves->getStart(),len,
                        (unsigned int)loc);
                */
    
                if (loc && len)
                    decode(loc, len);
            }
        }
        else if (desired_output==GOPCHOP_OUTPUT_PIPE && client_pipe)
        {
            // send entire packet to pipe

            len = packet->getLen();
            (void *)loc = mpeg2parser->bytesAvail(packet->getStart(), len);
            if (loc && client_pipe)
            {
                int written;
    
                // stuff things into the pipe
                while (len > 0)
                {
                    if ((written = write(*client_pipe, loc, len)) < 0)
                    {
                        perror("write");
                        exit(1);
                    }
                    loc += written;
                    len -= written;
                }
            }
        }

    }
    //printf("-- trying to flush pictures\n");
    /* throw in an end-of-stream to make sure libvo kicks out the last frame */
    if (desired_output==GOPCHOP_OUTPUT_LIBVO && output) {
#ifdef HAVE_MPEG2_FLUSH_PICTURE
        mpeg2_flush_picture(mpeg2dec);
        decode((uint8_t*)"",0);
#endif
        //decode((uint8_t*)seq_end_buf,sizeof(seq_end_buf));
    }
    //printf("-- done\n");
}

uint32_t get_GOP_selected()
{
    uint32_t GOP;

    GOP =
        (uint32_t)
        GTK_ADJUSTMENT(gtk_range_get_adjustment(GTK_RANGE(GOP_selector)))->
        value;
    return GOP;
}

void set_GOP_selected(uint32_t num)
{
    uint32_t max;
    uint32_t old;

    max = mpeg2parser->numGOPs();
    // correct our bounds
    if (num >= max)
        num = max - 1;
    if (num < 0)
        num = 0;

    // remember where we were
    old = get_GOP_selected();

    gtk_adjustment_set_value(GTK_ADJUSTMENT
                             (gtk_range_get_adjustment
                              (GTK_RANGE(GOP_selector))), num);

    // if we're in the same place, force a refresh, damnit
    if (num == old)
        gtk_adjustment_value_changed((GTK_ADJUSTMENT
                                      (gtk_range_get_adjustment
                                       (GTK_RANGE(GOP_selector)))));
}

void handle_rc_load()
{
    if (rc_load(PACKAGE,parsable_items))
        fprintf(stderr,"%s",_("Could not load rc file -- using defaults.\n"));

    /* synchronize the loaded/default values to the menus */
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menu_options_run_loop),options.run_loop);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(gtk_range_get_adjustment(GTK_RANGE(slider_run_speed))),
                             options.run_speed);
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menu_options_ignore_errors),options.ignore_errors);
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menu_options_auto_refresh),options.auto_refresh);
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menu_options_drop_orphaned_frames),options.drop_orphaned_frames);

}

void Usage(char *title)
{
    printf("Usage: %s [OPTIONS] [FILE]\n\
\n\
  -h, --help         This help\n\
  -v, --vo DRIVER    Choose internal output driver.  Use 'help' for a list.\n\
  -p, --pipe CMD     Use external command for output.  Recommended:\n\
                        'mplayer -nocache -novm -'\n\
  -s, --states       Report libmpeg2 states\n\
  FILE               MPEG2 to load on startup\n\
\n", title);

    exit(1);
}

void handle_args(int argc, char *argv[])
{
    static struct option long_options[] = {
        { "help",   0, NULL, 'h' }, // 0
        { "vo",     1, NULL, 'v' },
        { "pipe",   1, NULL, 'p' },
        { "states", 0, NULL, 's' },
        { 0,        0, 0,    0 }
    };

    /* do our option handling */
    while (1)
    {
        int c;
        int index;

        c = getopt_long(argc, argv, "hsv:p:", long_options, &index);

        if (c == -1)
            break; // done with args

        switch (c)
        {
            case 0:
                // got a matching long argument (with no short equiv)
                switch (index)
                {
                    default:
                        fprintf(stderr,
                                _("Unknown long argument index %d found!\n"),
                                index);
                }
                break;
            case 'h':
                Usage(argv[0]);
                break;
            case 's':
                opt_show_states=1;
                break;
            case 'v':
                opt_videoout=optarg;
                if (strcmp(opt_videoout,"help")==0)
                {
                    // display a list of libvo drivers
                    int i;
                    vo_driver_t const *drivers;

                    printf(_("Video output options:\n"));
                    drivers = vo_drivers();
                    for (i = 0; drivers[i].name; i++)
                        printf("\t%s\n",drivers[i].name);
                    exit(1);
                }
                else
                {
                    // verify selected libvo driver
                    if (!find_libvo_driver(opt_videoout))
                    {
                        fprintf(stderr,_("No such video driver: '%s'\n"),
                                opt_videoout);
                        exit(1);
                    }
                    desired_output=GOPCHOP_OUTPUT_LIBVO;
                }
                break;
            case 'p':
                opt_pipe=optarg;
                desired_output=GOPCHOP_OUTPUT_PIPE;
                break;
            default:
                fprintf(stderr, _("Unknown short argument '%c' found?!\n"), c);
                /* fall through to the Usage ... */
            case '?':
                Usage(argv[0]);
                break;
        }
    }
}

/*
 * There must be a way to hook up an event watcher to frame_window
 * directly.  Until I figure it out, this will have to do.  Overkill,
 * but it works.
 */
void gopchop_events(GdkEvent *event, gpointer data)
{
    if (frame_window &&
        event->type == GDK_VISIBILITY_NOTIFY &&
        ((GdkEventAny*)event)->window == frame_window &&
        ((GdkEventVisibility*)event)->state != GDK_VISIBILITY_FULLY_OBSCURED) {

        frame_draw();
    }

    gtk_main_do_event(event);
}

int main(int argc, char *argv[])
{


    /* deal with internationalization support */
#ifdef ENABLE_NLS
    setlocale(LC_ALL, "");
    bindtextdomain(PACKAGE, PACKAGE_LOCALE_DIR);
    textdomain(PACKAGE);
#endif
    gtk_set_locale();

    /* pass command line args to gtk (which may modify them) */
    gtk_init(&argc, &argv);
    /* handle our command line args */
    handle_args(argc, argv);

    /* setup the parsing object */
    mpeg2parser = new MPEG2Parser;

    /* set up all the gtk windows and widgets */
    setup_gtk_stuff();

    /* verify that we're working with largefile support
     * (report to Gtk window)
     */
    if (sizeof(off_t) < 8)
    {
        show_error(_("Type 'off_t' is less than 8 bytes.\n"
                     "Largefile support requires at least 8 bytes.\n"));
        gtk_widget_show(error_dialog);
    }

    /* load any saved options (prior to opening the first file) */
    handle_rc_load();

    /* attempt to load the filename mentioned on the command line 
     * (requires MPEG2Parser and GTK loaded
     */
    if (optind < argc)
        open_file(argv[optind]);

    /* start the event handler */
    gdk_event_handler_set(gopchop_events,NULL,NULL);
    gtk_main();
    return 0;
}


/* vi:set ai ts=4 sw=4 expandtab: */
